#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
Migration to IPA

Migrate users and groups from an LDAP server to IPA.

This performs an LDAP query against the remote server searching for
users and groups in a container. In order to migrate passwords you need
to bind as a user that can read the userPassword attribute on the remote
server. This is generally restricted to high-level admins such as
cn=Directory Manager in 389-ds (this is the default bind user).

The default user container is ou=People.

The default group container is ou=Groups.

Users and groups that already exist on the IPA server are skipped.

Two LDAP schemas define how group members are stored: RFC2307 and
RFC2307bis. RFC2307bis uses member and uniquemember to specify group
members, RFC2307 uses memberUid. The default schema is RFC2307bis.

The schema compat feature allows IPA to reformat data for systems that
do not support RFC2307bis. It is recommended that this feature is disabled
during migration to reduce system overhead. It can be re-enabled after
migration. To migrate with it enabled use the "--with-compat" option.

Migrated users do not have Kerberos credentials, they have only their
LDAP password. To complete the migration process, users need to go
to http://ipa.example.com/ipa/migration and authenticate using their
LDAP password in order to generate their Kerberos credentials.

Migration is disabled by default. Use the command ipa config-mod to
enable it:

 ipa config-mod --enable-migration=TRUE

If a base DN is not provided with --basedn then IPA will use either
the value of defaultNamingContext if it is set or the first value
in namingContexts set in the root of the remote LDAP server.

Users are added as members to the default user group. This can be a
time-intensive task so during migration this is done in a batch
mode for every 100 users. As a result there will be a window in which
users will be added to IPA but will not be members of the default
user group.

EXAMPLES:

 The simplest migration, accepting all defaults:
   ipa migrate-ds ldap://ds.example.com:389

 Specify the user and group container. This can be used to migrate user
 and group data from an IPA v1 server:
   ipa migrate-ds --user-container='cn=users,cn=accounts' \
       --group-container='cn=groups,cn=accounts' \
       ldap://ds.example.com:389

 Since IPA v2 server already contain predefined groups that may collide with
 groups in migrated (IPA v1) server (for example admins, ipausers), users
 having colliding group as their primary group may happen to belong to
 an unknown group on new IPA v2 server.
 Use --group-overwrite-gid option to overwrite GID of already existing groups
 to prevent this issue:
    ipa migrate-ds --group-overwrite-gid \
        --user-container='cn=users,cn=accounts' \
        --group-container='cn=groups,cn=accounts' \
        ldap://ds.example.com:389

 Migrated users or groups may have object class and accompanied attributes
 unknown to the IPA v2 server. These object classes and attributes may be
 left out of the migration process:
    ipa migrate-ds --user-container='cn=users,cn=accounts' \
       --group-container='cn=groups,cn=accounts' \
       --user-ignore-objectclass=radiusprofile \
       --user-ignore-attribute=radiusgroupname \
       ldap://ds.example.com:389

LOGGING

Migration will log warnings and errors to the Apache error log. This
file should be evaluated post-migration to correct or investigate any
issues that were discovered.

For every 100 users migrated an info-level message will be displayed to
give the current progress and duration to make it possible to track
the progress of migration.

If the log level is debug, either by setting debug = True in
/etc/ipa/default.conf or /etc/ipa/server.conf, then an entry will be printed
for each user added plus a summary when the default user group is
updated.
""")

register = Registry()


@register()
class migrate_ds(Command):
    __doc__ = _("Migrate users and groups from DS to IPA.")

    takes_args = (
        parameters.Str(
            'ldapuri',
            cli_name='ldap_uri',
            label=_(u'LDAP URI'),
            doc=_(u'LDAP URI of DS server to migrate from'),
        ),
        parameters.Password(
            'bindpw',
            cli_name='password',
            label=_(u'Password'),
            doc=_(u'bind password'),
        ),
    )
    takes_options = (
        parameters.DNParam(
            'binddn',
            required=False,
            cli_name='bind_dn',
            label=_(u'Bind DN'),
            default=DN(u'cn=directory manager'),
            autofill=True,
        ),
        parameters.DNParam(
            'usercontainer',
            cli_name='user_container',
            label=_(u'User container'),
            doc=_(u'DN of container for users in DS relative to base DN'),
            default=DN(u'ou=people'),
            autofill=True,
        ),
        parameters.DNParam(
            'groupcontainer',
            cli_name='group_container',
            label=_(u'Group container'),
            doc=_(u'DN of container for groups in DS relative to base DN'),
            default=DN(u'ou=groups'),
            autofill=True,
        ),
        parameters.Str(
            'userobjectclass',
            multivalue=True,
            cli_name='user_objectclass',
            label=_(u'User object class'),
            doc=_(u'Objectclasses used to search for user entries in DS'),
            default=(u'person',),
            autofill=True,
        ),
        parameters.Str(
            'groupobjectclass',
            multivalue=True,
            cli_name='group_objectclass',
            label=_(u'Group object class'),
            doc=_(u'Objectclasses used to search for group entries in DS'),
            default=(u'groupOfUniqueNames', u'groupOfNames'),
            autofill=True,
        ),
        parameters.Str(
            'userignoreobjectclass',
            required=False,
            multivalue=True,
            cli_name='user_ignore_objectclass',
            label=_(u'Ignore user object class'),
            doc=_(u'Objectclasses to be ignored for user entries in DS'),
            default=(),
            autofill=True,
        ),
        parameters.Str(
            'userignoreattribute',
            required=False,
            multivalue=True,
            cli_name='user_ignore_attribute',
            label=_(u'Ignore user attribute'),
            doc=_(u'Attributes to be ignored for user entries in DS'),
            default=(),
            autofill=True,
        ),
        parameters.Str(
            'groupignoreobjectclass',
            required=False,
            multivalue=True,
            cli_name='group_ignore_objectclass',
            label=_(u'Ignore group object class'),
            doc=_(u'Objectclasses to be ignored for group entries in DS'),
            default=(),
            autofill=True,
        ),
        parameters.Str(
            'groupignoreattribute',
            required=False,
            multivalue=True,
            cli_name='group_ignore_attribute',
            label=_(u'Ignore group attribute'),
            doc=_(u'Attributes to be ignored for group entries in DS'),
            default=(),
            autofill=True,
        ),
        parameters.Flag(
            'groupoverwritegid',
            cli_name='group_overwrite_gid',
            label=_(u'Overwrite GID'),
            doc=_(u'When migrating a group already existing in IPA domain overwrite the group GID and report as success'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'schema',
            required=False,
            cli_metavar="['RFC2307bis', 'RFC2307']",
            label=_(u'LDAP schema'),
            doc=_(u'The schema used on the LDAP server. Supported values are RFC2307 and RFC2307bis. The default is RFC2307bis'),
            default=u'RFC2307bis',
            autofill=True,
        ),
        parameters.Flag(
            'continue',
            required=False,
            label=_(u'Continue'),
            doc=_(u'Continuous operation mode. Errors are reported but the process continues'),
            default=False,
            autofill=True,
        ),
        parameters.DNParam(
            'basedn',
            required=False,
            cli_name='base_dn',
            label=_(u'Base DN'),
            doc=_(u'Base DN on remote LDAP server'),
        ),
        parameters.Flag(
            'compat',
            required=False,
            cli_name='with_compat',
            label=_(u'Ignore compat plugin'),
            doc=_(u'Allows migration despite the usage of compat plugin'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'cacertfile',
            required=False,
            cli_name='ca_cert_file',
            label=_(u'CA certificate'),
            doc=_(u'Load CA certificate of LDAP server from FILE'),
        ),
        parameters.Bool(
            'use_def_group',
            required=False,
            cli_name='use_default_group',
            label=_(u'Add to default group'),
            doc=_(u'Add migrated users without a group to a default group (default: true)'),
            default=True,
            autofill=True,
        ),
        parameters.Str(
            'scope',
            cli_metavar="['base', 'subtree', 'onelevel']",
            label=_(u'Search scope'),
            doc=_(u'LDAP search scope for users and groups: base, onelevel, or subtree. Defaults to onelevel'),
            default=u'onelevel',
            autofill=True,
        ),
        parameters.Str(
            'exclude_groups',
            required=False,
            multivalue=True,
            doc=_(u'groups to exclude from migration'),
            default=(),
            autofill=True,
        ),
        parameters.Str(
            'exclude_users',
            required=False,
            multivalue=True,
            doc=_(u'users to exclude from migration'),
            default=(),
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'result',
            dict,
            doc=_(u'Lists of objects migrated; categorized by type.'),
        ),
        output.Output(
            'failed',
            dict,
            doc=_(u'Lists of objects that could not be migrated; categorized by type.'),
        ),
        output.Output(
            'enabled',
            bool,
            doc=_(u'False if migration mode was disabled.'),
        ),
        output.Output(
            'compat',
            bool,
            doc=_(u'False if migration fails because the compatibility plug-in is enabled.'),
        ),
    )



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Sudo tests
"""

from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
import ipatests.test_webui.data_sudo as sudo
import ipatests.test_webui.data_netgroup as netgroup
import ipatests.test_webui.data_user as user
import ipatests.test_webui.data_group as group
import ipatests.test_webui.data_hostgroup as hostgroup
from ipatests.test_webui.test_host import host_tasks, ENTITY as HOST_ENTITY
import pytest


@pytest.mark.tier1
class test_sudo(UI_driver):

    @screenshot
    def test_crud(self):
        """
        Basic CRUD: sudo
        """
        self.init_app()
        self.basic_crud(sudo.RULE_ENTITY, sudo.RULE_DATA)
        self.basic_crud(sudo.CMDENTITY, sudo.CMD_DATA)
        self.basic_crud(sudo.CMDGROUP_ENTITY, sudo.CMDGROUP_DATA,
                        default_facet=sudo.CMDGROUP_DEF_FACET)

    @screenshot
    def test_mod(self):
        """
        Mod: sudo
        """
        self.init_app()
        host = host_tasks()
        host.driver = self.driver
        host.config = self.config
        host.prep_data()
        host.prep_data2()

        self.add_record(netgroup.ENTITY, netgroup.DATA2)

        self.add_record(user.ENTITY, user.DATA)
        self.add_record(user.ENTITY, user.DATA2, navigate=False)

        self.add_record(group.ENTITY, group.DATA)
        self.add_record(group.ENTITY, group.DATA2, navigate=False)

        self.add_record(HOST_ENTITY, host.data)
        self.add_record(HOST_ENTITY, host.data2, navigate=False)

        self.add_record(hostgroup.ENTITY, hostgroup.DATA)
        self.add_record(hostgroup.ENTITY, hostgroup.DATA2, navigate=False)

        self.add_record(sudo.CMDENTITY, sudo.CMD_DATA)
        self.add_record(sudo.CMDENTITY, sudo.CMD_DATA2, navigate=False)

        self.add_record(sudo.CMDGROUP_ENTITY, sudo.CMDGROUP_DATA)
        self.add_record(sudo.CMDGROUP_ENTITY, sudo.CMDGROUP_DATA2, navigate=False)

        self.add_record(sudo.RULE_ENTITY, sudo.RULE_DATA)
        self.navigate_to_record(sudo.RULE_PKEY, entity=sudo.RULE_ENTITY)

        tables = [
            ['memberuser_user', [user.PKEY, user.PKEY2], ],
            ['memberuser_group', [group.PKEY, group.PKEY2], ],
            ['memberhost_host', [host.pkey, host.pkey2], ],
            ['memberhost_hostgroup', [hostgroup.PKEY, hostgroup.PKEY2], ],
            ['memberallowcmd_sudocmd', [sudo.CMD_PKEY, sudo.CMD_PKEY2], ],
            ['memberallowcmd_sudocmdgroup', [sudo.CMD_GROUP_PKEY, sudo.CMD_GROUP_PKEY2], ],
            ['memberdenycmd_sudocmd', [sudo.CMD_PKEY, sudo.CMD_PKEY2], ],
            ['memberdenycmd_sudocmdgroup', [sudo.CMD_GROUP_PKEY, sudo.CMD_GROUP_PKEY2], ],
            ['ipasudorunas_user', ['admin'], ],
            ['ipasudorunas_group', ['editors', 'admins'], ],
            ['ipasudorunasgroup_group', ['editors', 'admins'], ],
        ]

        categories = [
            'usercategory',
            'hostcategory',
            'cmdcategory',
            'ipasudorunasusercategory',
            'ipasudorunasgroupcategory',
        ]

        no_cats = [
            'memberdenycmd_sudocmd',
            'memberdenycmd_sudocmdgroup',
        ]

        self.mod_rule_tables(tables, categories, no_cats)

        # cleanup
        # -------
        self.delete(sudo.RULE_ENTITY, [sudo.RULE_DATA])
        self.delete(user.ENTITY, [user.DATA, user.DATA2])
        self.delete(group.ENTITY, [group.DATA, group.DATA2])
        self.delete(HOST_ENTITY, [host.data, host.data2])
        self.delete(hostgroup.ENTITY, [hostgroup.DATA, hostgroup.DATA2])
        self.delete(sudo.CMDENTITY, [sudo.CMD_DATA, sudo.CMD_DATA2])
        self.delete(sudo.CMDGROUP_ENTITY, [sudo.CMDGROUP_DATA, sudo.CMDGROUP_DATA2])

    @screenshot
    def test_actions(self):
        """
        Test sudo rule actions
        """
        self.init_app()

        self.add_record(sudo.RULE_ENTITY, sudo.RULE_DATA)
        self.navigate_to_record(sudo.RULE_PKEY)

        self.disable_action()
        self.enable_action()
        self.delete_action(sudo.RULE_ENTITY, sudo.RULE_PKEY)



# Authors:
#   Rob Crittenden <rcritten@redhat.com>
#
# Copyright (C) 2009,2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
Test the `ipaserver/plugins/cert.py` module against a RA.
"""
from __future__ import print_function, absolute_import

import base64
import os
import unittest

import pytest
import six
from ipalib import api
from ipalib import errors
from ipaplatform.paths import paths
from ipapython.certdb import NSSDatabase
from ipapython.dn import DN
from ipapython.ipautil import run
from ipatests.test_xmlrpc.testcert import subject_base
from ipatests.test_xmlrpc.xmlrpc_test import XMLRPC_test

if six.PY3:
    unicode = str

# So we can save the cert from issuance and compare it later
cert = None
newcert = None
sn = None

_DOMAIN = api.env.domain
_EXP_CRL_URI = ''.join(['http://ipa-ca.', _DOMAIN, '/ipa/crl/MasterCRL.bin'])
_EXP_OCSP_URI = ''.join(['http://ipa-ca.', _DOMAIN, '/ca/ocsp'])

def is_db_configured():
    """
    Raise an exception if we are testing against lite-server and the
    developer cert database is configured.
    """
    aliasdir = api.env.dot_ipa + os.sep + 'alias' + os.sep + '.pwd'

    if (api.env.xmlrpc_uri == u'http://localhost:8888/ipa/xml' and
       not os.path.isfile(aliasdir)):
        raise unittest.SkipTest('developer CA not configured in %s' % aliasdir)

# Test setup
#
# This test needs a configured CA behind it in order to work properly
#
# To test against Apache directly then no changes are required. Just be
# sure the xmlrpc_uri in ~/.ipa/default.conf points to Apache.
#
# To test against Dogtag CA in the lite-server:
#
# - Copy the 3 NSS db files from /var/lib/ipa/radb to ~/.ipa/alias
# - Copy /var/lib/ipa/radb/pwdfile.txt to  ~/.ipa/alias/.pwd.
# - Change ownership of these files to be readable by you.
#
# The API tested depends on the value of ~/.ipa/default/ra_plugin when
# running as the lite-server.

class BaseCert(XMLRPC_test):
    host_fqdn = u'ipatestcert.%s' % api.env.domain
    service_princ = u'test/%s@%s' % (host_fqdn, api.env.realm)

    @classmethod
    def setup_class(cls):
        super(BaseCert, cls).setup_class()

        if 'cert_request' not in api.Command:
            raise unittest.SkipTest('cert_request not registered')
        if 'cert_show' not in api.Command:
            raise unittest.SkipTest('cert_show not registered')

        is_db_configured()

    def setup(self):
        self.nssdb = NSSDatabase()
        secdir = self.nssdb.secdir
        self.reqfile = os.path.join(secdir, "test.csr")
        self.certfile = os.path.join(secdir, "cert.crt")
        # Create our temporary NSS database
        self.nssdb.create_db()
        self.subject = DN(('CN', self.host_fqdn), subject_base())

    def teardown(self):
        self.nssdb.close()  # remove tempdir

    def generateCSR(self, subject):
        self.nssdb.run_certutil([
            "-R", "-s", subject,
            "-o", self.reqfile,
            "-z", paths.GROUP,
            "-a",
        ])
        with open(self.reqfile, "rb") as f:
            return f.read().decode('ascii')


@pytest.mark.tier1
class test_cert(BaseCert):

    @classmethod
    def setup_class(cls):
        super(test_cert, cls).setup_class()

    """
    Test the `cert` plugin.
    """

    def test_0001_cert_add(self):
        """
        Test the `xmlrpc.cert_request` method without --add.

        This should fail because the service principal doesn't exist
        """
        # First create the host that will use this policy
        assert 'result' in api.Command['host_add'](self.host_fqdn, force=True)

        csr = self.generateCSR(str(self.subject))
        with pytest.raises(errors.NotFound):
            api.Command['cert_request'](csr, principal=self.service_princ)

    def test_0002_cert_add(self):
        """
        Test the `xmlrpc.cert_request` method with --add.
        """
        # Our host should exist from previous test
        global cert, sn

        csr = self.generateCSR(str(self.subject))
        res = api.Command['cert_request'](csr, principal=self.service_princ, add=True)['result']
        assert DN(res['subject']) == self.subject
        assert 'cacn' in res
        # save the cert for the service_show/find tests
        cert = res['certificate'].encode('ascii')
        # save cert's SN for URI test
        sn = res['serial_number']

    def test_0003_service_show(self):
        """
        Verify that service-show has the right certificate using service-show.
        """
        res = api.Command['service_show'](self.service_princ)['result']
        assert base64.b64encode(res['usercertificate'][0]) == cert

    def test_0004_service_find(self):
        """
        Verify that service-find has the right certificate using service-find.
        """
        # Assume there is only one service
        res = api.Command['service_find'](self.service_princ)['result']
        assert base64.b64encode(res[0]['usercertificate'][0]) == cert

    def test_0005_cert_uris(self):
        """Test URI details and OCSP-URI in certificate.

        See https://fedorahosted.org/freeipa/ticket/5881
        """
        result = api.Command.cert_show(sn, out=unicode(self.certfile))
        with open(self.certfile, "rb") as f:
            pem_cert = f.read().decode('ascii')
        result = run([paths.OPENSSL, 'x509', '-text'],
                     stdin=pem_cert, capture_output=True)
        assert _EXP_CRL_URI in result.output
        assert _EXP_OCSP_URI in result.output

    def test_0006_cert_renew(self):
        """
        Issue a new certificate for a service
        """
        global newcert

        csr = self.generateCSR(str(self.subject))
        res = api.Command['cert_request'](csr, principal=self.service_princ)['result']
        assert DN(res['subject']) == self.subject
        # save the cert for the service_show/find tests
        newcert = res['certificate'].encode('ascii')

    def test_0007_service_show(self):
        """
        Verify the new certificate with service-show.
        """
        res = api.Command['service_show'](self.service_princ)['result']

        # Both the old and the new certs should be listed as certificates now
        certs_encoded = (
            base64.b64encode(usercert) for usercert in res['usercertificate']
        )
        assert set(certs_encoded) == set([cert, newcert])

    def test_0008_cert_show(self):
        """
        Verify that cert-show shows CA of the certificate without --all
        """
        res = api.Command['cert_show'](sn)['result']
        assert 'cacn' in res
        assert 'valid_not_before' in res
        assert 'valid_not_after' in res

    def test_0009_cert_find(self):
        """
        Verify that cert-find shows CA of the certificate without --all
        """
        res = api.Command['cert_find'](min_serial_number=sn,
                                       max_serial_number=sn)['result'][0]
        assert 'cacn' in res
        assert 'valid_not_before' in res
        assert 'valid_not_after' in res

    def test_00010_san_in_cert(self):
        """
        Test if SAN extension is automatically added with default profile.
        """
        csr = self.generateCSR(str(self.subject))
        res = api.Command[
            'cert_request'](csr, principal=self.service_princ)['result']
        assert 'san_dnsname' in res

    def test_00011_emails_are_valid(self):
        """
        Verify the different scenarios when checking if any email addr
        from DN or SAN extension does not appear in ldap entry.
        """

        from ipaserver.plugins.cert import _emails_are_valid
        email_addrs = [u'any@EmAiL.CoM']
        result = _emails_are_valid(email_addrs, [u'any@email.com'])
        assert True == result, result

        email_addrs = [u'any@EmAiL.CoM']
        result = _emails_are_valid(email_addrs, [u'any@email.com',
                                                 u'another@email.com'])
        assert True == result, result

        result = _emails_are_valid([], [u'any@email.com'])
        assert True == result, result

        email_addrs = [u'invalidEmailAddress']
        result = _emails_are_valid(email_addrs, [])
        assert False == result, result

    def test_99999_cleanup(self):
        """
        Clean up cert test data
        """
        # Now clean things up
        api.Command['host_del'](self.host_fqdn)

        # Verify that the service is gone
        res = api.Command['service_find'](self.service_princ)
        assert res['count'] == 0


@pytest.mark.tier1
class test_cert_find(XMLRPC_test):

    @classmethod
    def setup_class(cls):
        super(test_cert_find, cls).setup_class()

        if 'cert_find' not in api.Command:
            raise unittest.SkipTest('cert_find not registered')

        if api.env.ra_plugin != 'dogtag':
            raise unittest.SkipTest('cert_find for dogtag CA only')

        is_db_configured()

    """
    Test the `cert-find` command.
    """
    short = api.env.host.split('.')[0]

    def test_0001_find_all(self):
        """
        Search for all certificates.

        We don't know how many we'll get but there should be at least 10
        by default.
        """
        res = api.Command['cert_find']()
        assert 'count' in res and res['count'] >= 10

    def test_0002_find_CA(self):
        """
        Search for the CA certificate.
        """
        res = api.Command['cert_find'](subject=u'Certificate Authority')
        assert 'count' in res and res['count'] == 1

    def test_0003_find_OCSP(self):
        """
        Search for the OCSP certificate.
        """
        res = api.Command['cert_find'](subject=u'OCSP Subsystem')
        assert 'count' in res
        assert res['count'], "No OSCP certificate found"

    def test_0004_find_this_host(self):
        """
        Find all certificates for this IPA server
        """
        res = api.Command['cert_find'](subject=api.env.host)
        assert 'count' in res and res['count'] > 1

    def test_0005_find_this_host_exact(self):
        """
        Find all certificates for this IPA server (exact)
        """
        res = api.Command['cert_find'](subject=api.env.host, exactly=True)
        assert 'count' in res and res['count'] > 1

    def test_0006_find_this_short_host_exact(self):
        """
        Find all certificates for this IPA server short name (exact)
        """
        res = api.Command['cert_find'](subject=self.short, exactly=True)
        assert 'count' in res and res['count'] == 0

    # tests 0007 to 0016 removed

    def test_0017_find_by_issuedon(self):
        """
        Find all certificates issued since 2008
        """
        res = api.Command['cert_find'](issuedon_from=u'2008-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 10

    def test_0018_find_through_issuedon(self):
        """
        Find all certificates issued through 2008
        """
        res = api.Command['cert_find'](issuedon_to=u'2008-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 0

    def test_0019_find_notvalid_before(self):
        """
        Find all certificates valid not before 2008
        """
        res = api.Command['cert_find'](validnotbefore_from=u'2008-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 10

    def test_0020_find_notvalid_before(self):
        """
        Find all certificates valid not before to 2100
        """
        res = api.Command['cert_find'](validnotbefore_to=u'2100-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 10

    def test_0021_find_notvalid_before(self):
        """
        Find all certificates valid not before 2100
        """
        res = api.Command['cert_find'](validnotbefore_from=u'2100-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 0

    def test_0022_find_notvalid_before(self):
        """
        Find all certificates valid not before to 2008
        """
        res = api.Command['cert_find'](validnotbefore_to=u'2008-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 0

    def test_0023_find_notvalid_after(self):
        """
        Find all certificates valid not after 2008
        """
        res = api.Command['cert_find'](validnotafter_from=u'2008-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 10

    def test_0024_find_notvalid_after(self):
        """
        Find all certificates valid not after to 2100
        """
        res = api.Command['cert_find'](validnotafter_to=u'2100-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 10

    def test_0025_find_notvalid_after(self):
        """
        Find all certificates valid not after 2100
        """
        res = api.Command['cert_find'](validnotafter_from=u'2100-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 0

    def test_0026_find_notvalid_after(self):
        """
        Find all certificates valid not after to 2008
        """
        res = api.Command['cert_find'](validnotafter_to=u'2008-01-01',
                                       sizelimit=10)
        assert 'count' in res and res['count'] == 0

    def test_0027_sizelimit_zero(self):
        """
        Search with a sizelimit of 0
        """
        count_all = api.Command['cert_find']()['count']
        res = api.Command['cert_find'](sizelimit=0)
        assert 'count' in res and res['count'] == count_all

    def test_0028_find_negative_size(self):
        """
        Search with a negative sizelimit
        """
        with pytest.raises(errors.ValidationError):
            api.Command['cert_find'](sizelimit=-100)

    def test_0029_search_for_notfound(self):
        """
        Search for a host that isn't there.
        """
        res = api.Command['cert_find'](subject=u'notfound')
        assert 'count' in res and res['count'] == 0

    def test_0030_search_for_testcerts(self):
        """
        Search for certs created in other tests
        """
        res = api.Command['cert_find'](subject=u'ipatestcert.%s' % api.env.domain)
        assert 'count' in res and res['count'] >= 1

    def test_0031_search_on_invalid_date(self):
        """
        Search using invalid date format
        """
        with pytest.raises(errors.ConversionError):
            api.Command['cert_find'](issuedon_from=u'xyz')


@pytest.mark.tier1
class test_cert_revocation(BaseCert):

    @classmethod
    def setup_class(cls):
        super(test_cert_revocation, cls).setup_class()

    # create CSR, request cert, revoke cert, check cert attributes
    def revoke_cert(self, reason):
        # add host
        assert 'result' in api.Command['host_add'](self.host_fqdn, force=True)

        # generate CSR, request certificate, obtain serial number
        self.csr = self.generateCSR(str(self.subject))
        res = api.Command['cert_request'](self.csr,
                                          principal=self.service_princ,
                                          add=True, all=True)['result']
        serial_number = res['serial_number']

        # revoke created certificate
        assert 'result' in api.Command['cert_revoke'](
            serial_number, revocation_reason=reason)

        # verify that certificate is revoked with correct reason
        res2 = api.Command['cert_show'](serial_number, all=True)['result']
        assert res2['revoked']
        assert res2['revocation_reason'] == reason

        # remove host
        assert 'result' in api.Command['host_del'](self.host_fqdn)

    def test_revoke_with_reason_0(self):
        self.revoke_cert(0)

    def test_revoke_with_reason_1(self):
        self.revoke_cert(1)

    def test_revoke_with_reason_2(self):
        self.revoke_cert(2)

    def test_revoke_with_reason_3(self):
        self.revoke_cert(3)

    def test_revoke_with_reason_4(self):
        self.revoke_cert(4)

    def test_revoke_with_reason_5(self):
        self.revoke_cert(5)

    def test_revoke_with_reason_6(self):
        self.revoke_cert(6)

    def test_revoke_with_reason_8(self):
        self.revoke_cert(8)

    def test_revoke_with_reason_9(self):
        self.revoke_cert(9)

    def test_revoke_with_reason_10(self):
        self.revoke_cert(10)



# coding: utf-8
#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#
from contextlib import contextmanager
import six

from ipalib import api
from ipatests.util import MockLDAP

trust_container_dn = "cn=ad,cn=trusts,{basedn}".format(
    basedn=api.env.basedn)
trust_container_add = dict(
    objectClass=[b"nsContainer", b"top"]
    )

smb_cont_dn = "{cifsdomains},{basedn}".format(
    cifsdomains=api.env.container_cifsdomains,
    basedn=api.env.basedn)
smb_cont_add = dict(
    objectClass=[b"nsContainer", b"top"]
    )


def create_mock_trust_containers():
    with MockLDAP() as ldap:
        ldap.add_entry(trust_container_dn, trust_container_add)
        ldap.add_entry(smb_cont_dn, smb_cont_add)


def remove_mock_trust_containers():
    with MockLDAP() as ldap:
        ldap.del_entry(trust_container_dn)
        ldap.del_entry(smb_cont_dn)


@contextmanager
def mocked_trust_containers():
    """Mocked trust containers

    Provides containers for the RPC tests:
    cn=ad,cn=trusts,BASEDN
    cn=ad,cn=etc,BASEDN

    Upon exiting, it tries to remove the container entries.
    If the user of the context manager failed to remove
    all child entries, exiting the context manager will fail.
    """
    create_mock_trust_containers()
    try:
        yield
    finally:
        remove_mock_trust_containers()

def get_range_dn(name):
    format_str = "cn={name},cn=ranges,cn=etc,{basedn}"
    data = dict(name=name, basedn=api.env.basedn)
    return format_str.format(**data)


def get_trust_dn(name):
    format_str = "cn={name},cn=ad,cn=trusts,{basedn}"
    data = dict(name=name, basedn=api.env.basedn)
    return format_str.format(**data)


def encode_mockldap_value(value):
    value = str(value)
    if six.PY3:
        return value.encode('utf-8')
    else:
        return value


def get_trusted_dom_range_dict(name, base_id, size, rangetype, base_rid, sid):
    return dict(
        objectClass=[b"ipaIDrange", b"ipatrustedaddomainrange"],
        ipaBaseID=encode_mockldap_value("{base_id}".format(base_id=base_id)),
        ipaBaseRID=encode_mockldap_value("{base_rid}".format(base_rid=base_rid)),
        ipaIDRangeSize=encode_mockldap_value("{size}".format(size=size)),
        ipaNTTrustedDomainSID=encode_mockldap_value("{sid}".format(sid=sid)),
        ipaRangeType=encode_mockldap_value("{rangetype}".format(rangetype=rangetype)),
        )


def get_trusted_dom_dict(name, sid):
    return dict(
        objectClass=[b"ipaNTTrustedDomain", b"ipaIDobject", b"top"],
        ipaNTFlatName=encode_mockldap_value(name.split('.')[0].upper()),
        ipaNTTrustedDomainSID=encode_mockldap_value(sid),
        ipaNTSIDBlacklistIncoming=b'S-1-0',
        ipaNTTrustPartner=encode_mockldap_value('{name}.mock'.format(name=name)),
        ipaNTTrustType=b'2',
        ipaNTTrustDirection=b'3',
        ipaNTTrustAttributes=b'8',
        )



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
Plugins not accessible directly through the CLI, commands used internally
""")

register = Registry()


@register()
class i18n_messages(Command):
    NO_CLI = True

    has_output = (
        output.Output(
            'messages',
            dict,
            doc=_(u'Dict of I18N messages'),
        ),
    )


@register()
class json_metadata(Command):
    __doc__ = _("Export plugin meta-data for the webUI.")

    NO_CLI = True

    takes_args = (
        parameters.Str(
            'objname',
            required=False,
            doc=_(u'Name of object to export'),
        ),
        parameters.Str(
            'methodname',
            required=False,
            doc=_(u'Name of method to export'),
        ),
    )
    takes_options = (
        parameters.Str(
            'object',
            required=False,
            doc=_(u'Name of object to export'),
        ),
        parameters.Str(
            'method',
            required=False,
            doc=_(u'Name of method to export'),
        ),
        parameters.Str(
            'command',
            required=False,
            doc=_(u'Name of command to export'),
        ),
    )
    has_output = (
        output.Output(
            'objects',
            dict,
            doc=_(u'Dict of JSON encoded IPA Objects'),
        ),
        output.Output(
            'methods',
            dict,
            doc=_(u'Dict of JSON encoded IPA Methods'),
        ),
        output.Output(
            'commands',
            dict,
            doc=_(u'Dict of JSON encoded IPA Commands'),
        ),
    )



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

import errno
import json
import logging
import os
import sys
import tempfile
import types
import zipfile

from cryptography import x509 as crypto_x509

import six

from ipaclient.frontend import ClientCommand, ClientMethod
from ipalib import errors, parameters, plugable
from ipalib.constants import USER_CACHE_PATH
from ipalib.errors import SchemaUpToDate
from ipalib.frontend import Object
from ipalib.output import Output
from ipalib.parameters import DefaultFrom, Flag, Password, Str
from ipapython import ipautil
from ipapython.ipautil import fsdecode
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

# pylint: disable=no-name-in-module, import-error
if six.PY3:
    from collections.abc import Mapping, Sequence
else:
    from collections import Mapping, Sequence
# pylint: enable=no-name-in-module, import-error

logger = logging.getLogger(__name__)

FORMAT = '1'

if six.PY3:
    unicode = str

_TYPES = {
    'DN': DN,
    'DNSName': DNSName,
    'Principal': unicode,
    'NoneType': type(None),
    'Sequence': Sequence,
    'bool': bool,
    'dict': dict,
    'int': int,
    'list': list,
    'tuple': tuple,
    'unicode': unicode,
    'Certificate': crypto_x509.Certificate,
}

_PARAMS = {
    'Decimal': parameters.Decimal,
    'DN': parameters.DNParam,
    'DNSName': parameters.DNSNameParam,
    'Principal': parameters.Principal,
    'bool': parameters.Bool,
    'bytes': parameters.Bytes,
    'datetime': parameters.DateTime,
    'dict': parameters.Dict,
    'int': parameters.Int,
    'str': parameters.Str,
    'Certificate': parameters.Certificate,
}


def json_default(obj):
    if isinstance(obj, bytes):
        return obj.decode('utf-8')
    raise TypeError


class _SchemaCommand(ClientCommand):
    pass


class _SchemaMethod(ClientMethod):
    @property
    def obj_name(self):
        return self.api.Object[self.obj_full_name].name

    @property
    def obj_version(self):
        return self.api.Object[self.obj_full_name].version


class _SchemaObject(Object):
    pass


class _SchemaPlugin:
    bases = None
    schema_key = None

    def __init__(self, schema, full_name):
        self.name, _slash, self.version = full_name.partition('/')
        self.full_name = full_name
        self._schema = schema
        self._class = None

    @property
    def doc(self):
        if self._class is not None:
            return self._class.doc
        else:
            schema = self._schema[self.schema_key][self.full_name]
            try:
                return schema['doc']
            except KeyError:
                return None

    @property
    def summary(self):
        if self._class is not None:
            return self._class.summary
        else:
            halp = self._schema[self.schema_key].get_help(self.full_name)
            try:
                return halp['summary']
            except KeyError:
                return u'<%s>' % self.full_name

    def _create_default_from(self, api, name, keys):
        cmd_name = self.full_name

        def get_default(*args):
            kw = dict(zip(keys, args))
            result = api.Command.command_defaults(
                unicode(cmd_name),
                params=[unicode(name)],
                kw=kw,
            )['result']
            return result.get(name)

        if keys:
            def callback(*args):
                return get_default(*args)
        else:
            def callback():
                return get_default()

        callback.__name__ = '{0}_{1}_default'.format(self.name, name)

        return DefaultFrom(callback, *keys)

    def _create_param(self, api, schema):
        name = str(schema['name'])
        type_name = str(schema['type'])
        sensitive = schema.get('sensitive', False)

        if type_name == 'str' and sensitive:
            cls = Password
            sensitive = False
        elif (type_name == 'bool' and
                'default' in schema and schema['default'][0] == u'False' and
                not schema.get('alwaysask', False)):
            cls = Flag
            del schema['default']
        else:
            try:
                cls = _PARAMS[type_name]
            except KeyError:
                cls = Str

        kwargs = {}
        default = None

        for key, value in schema.items():
            if key in ('alwaysask',
                       'doc',
                       'label',
                       'multivalue',
                       'no_convert',
                       'option_group',
                       'required'):
                kwargs[key] = value
            elif key in ('cli_metavar',
                         'cli_name'):
                kwargs[key] = str(value)
            elif key == 'confirm':
                kwargs[key] = value
            elif key == 'default':
                default = value
            elif key == 'default_from_param':
                keys = tuple(str(k) for k in value)
                kwargs['default_from'] = (
                    self._create_default_from(api, name, keys))
            elif key in ('exclude',
                         'include'):
                kwargs[key] = tuple(str(v) for v in value)

        if default is not None:
            tmp = cls(name, **dict(kwargs, no_convert=False))
            if tmp.multivalue:
                default = tuple(tmp._convert_scalar(d) for d in default)
            else:
                default = tmp._convert_scalar(default[0])
            kwargs['default'] = default

        if 'default' in kwargs or 'default_from' in kwargs:
            kwargs['autofill'] = not kwargs.pop('alwaysask', False)

        param = cls(name, **kwargs)

        if sensitive:
            object.__setattr__(param, 'password', True)

        return param

    def _create_class(self, api, schema):
        class_dict = {}

        class_dict['name'] = str(schema['name'])
        class_dict['version'] = str(schema['version'])
        class_dict['full_name'] = str(schema['full_name'])
        if 'doc' in schema:
            class_dict['doc'] = schema['doc']
        if 'topic_topic' in schema:
            class_dict['topic'] = str(schema['topic_topic']).partition('/')[0]
        else:
            class_dict['topic'] = None

        class_dict['takes_params'] = tuple(self._create_param(api, s)
                                           for s in schema.get('params', []))

        return self.name, self.bases, class_dict

    def __call__(self, api):
        if self._class is None:
            schema = self._schema[self.schema_key][self.full_name]
            name, bases, class_dict = self._create_class(api, schema)
            self._class = type(name, bases, class_dict)

        return self._class(api)


class _SchemaCommandPlugin(_SchemaPlugin):
    bases = (_SchemaCommand,)
    schema_key = 'commands'

    @property
    def topic(self):
        if self._class is not None:
            return self._class.topic
        else:
            halp = self._schema[self.schema_key].get_help(self.full_name)
            try:
                return str(halp['topic_topic']).partition('/')[0]
            except KeyError:
                return None

    @property
    def NO_CLI(self):
        if self._class is not None:
            return self._class.NO_CLI
        else:
            halp = self._schema[self.schema_key].get_help(self.full_name)
            return 'cli' in halp.get('exclude', [])

    def _create_output(self, api, schema):
        if schema.get('multivalue', False):
            type_type = (tuple, list)
            if not schema.get('required', True):
                type_type = type_type + (type(None),)
        else:
            try:
                type_type = _TYPES[schema['type']]
            except KeyError:
                type_type = None
            else:
                if not schema.get('required', True):
                    type_type = (type_type, type(None))

        kwargs = {}
        kwargs['type'] = type_type

        if 'doc' in schema:
            kwargs['doc'] = schema['doc']

        if schema.get('no_display', False):
            kwargs['flags'] = ('no_display',)

        return Output(str(schema['name']), **kwargs)

    def _create_class(self, api, schema):
        name, bases, class_dict = (
            super(_SchemaCommandPlugin, self)._create_class(api, schema))

        if 'obj_class' in schema or 'attr_name' in schema:
            bases = (_SchemaMethod,)

        if 'obj_class' in schema:
            class_dict['obj_full_name'] = str(schema['obj_class'])
        if 'attr_name' in schema:
            class_dict['attr_name'] = str(schema['attr_name'])
        if 'exclude' in schema and u'cli' in schema['exclude']:
            class_dict['NO_CLI'] = True

        args = set(str(s['name']) for s in schema['params']
                   if s.get('positional', s.get('required', True)))
        class_dict['takes_args'] = tuple(
            p for p in class_dict['takes_params'] if p.name in args)
        class_dict['takes_options'] = tuple(
            p for p in class_dict['takes_params'] if p.name not in args)
        del class_dict['takes_params']

        class_dict['has_output'] = tuple(
            self._create_output(api, s) for s in schema['output'])

        return name, bases, class_dict


class _SchemaObjectPlugin(_SchemaPlugin):
    bases = (_SchemaObject,)
    schema_key = 'classes'


class _SchemaNameSpace(Mapping):

    def __init__(self, schema, name):
        self.name = name
        self._schema = schema

    def __getitem__(self, key):
        try:
            return self._schema.read_namespace_member(self.name, key)
        except KeyError:
            raise KeyError(key)

    def __iter__(self):
        for key in self._schema.iter_namespace(self.name):
            yield key

    def __len__(self):
        return len(list(self._schema.iter_namespace(self.name)))

    def get_help(self, key):
        try:
            return self._schema.get_help(self.name, key)
        except KeyError:
            raise KeyError(key)


class NotAvailable(Exception):
    pass


class Schema:
    """
    Store and provide schema for commands and topics

    Create api instance
    >>> from ipalib import api
    >>> api.bootstrap(context='cli')
    >>> api.finalize()

    Get schema object
    >>> m = Schema(api)

    From now on we can access schema for commands stored in cache
    >>> m['commands'][u'ping'][u'doc']
    u'Ping a remote server.'

    >>> m['topics'][u'ping'][u'doc']
    u'Ping the remote IPA server to ...'

    """
    namespaces = {'classes', 'commands', 'topics'}
    _DIR = os.path.join(USER_CACHE_PATH, 'ipa', 'schema', FORMAT)

    def __init__(self, client, fingerprint=None):
        self._dict = {}
        self._namespaces = {}
        self._help = None

        for ns in self.namespaces:
            self._dict[ns] = {}
            self._namespaces[ns] = _SchemaNameSpace(self, ns)

        ttl = None
        read_failed = False

        if fingerprint is not None:
            try:
                self._read_schema(fingerprint)
            except Exception as e:
                # Failed to read the schema from cache. There may be a lot of
                # causes and not much we can do about it. Just ensure we will
                # ignore the cache and fetch the schema from server.
                logger.warning("Failed to read schema: %s", e)
                fingerprint = None
                read_failed = True

        if fingerprint is None:
            fingerprint, ttl = self._fetch(client, ignore_cache=read_failed)
            self._help = self._generate_help(self._dict)
            try:
                self._write_schema(fingerprint)
            except Exception as e:
                logger.warning("Failed to write schema: %s", e)

        self.fingerprint = fingerprint
        self.ttl = ttl

    def _fetch(self, client, ignore_cache=False):
        if not client.isconnected():
            client.connect(verbose=False)

        fps = []
        if not ignore_cache:
            try:
                fps = [fsdecode(f) for f in os.listdir(self._DIR)]
            except EnvironmentError:
                pass

        kwargs = {u'version': u'2.170'}
        if fps:
            kwargs[u'known_fingerprints'] = fps
        try:
            schema = client.forward(u'schema', **kwargs)['result']
        except errors.CommandError:
            raise NotAvailable()

        try:
            fp = schema['fingerprint']
            ttl = schema.pop('ttl')
            schema.pop('version')

            for key, value in schema.items():
                if key in self.namespaces:
                    value = {m['full_name']: m for m in value}
                self._dict[key] = value
        except KeyError as e:
            logger.warning("Failed to fetch schema: %s", e)
            raise NotAvailable()

        return (fp, ttl,)

    def _read_schema(self, fingerprint):
        # It's more efficient to read zip file members at once than to open
        # the zip file a couple of times, see #6690.
        filename = os.path.join(self._DIR, fingerprint)
        with zipfile.ZipFile(filename, 'r') as schema:
            for name in schema.namelist():
                ns, _slash, key = name.partition('/')
                if ns in self.namespaces:
                    self._dict[ns][key] = schema.read(name)
                elif name == '_help':
                    self._help = schema.read(name)

    def __getitem__(self, key):
        try:
            return self._namespaces[key]
        except KeyError:
            return self._dict[key]

    def _generate_help(self, schema):
        halp = {}

        for namespace in ('commands', 'topics'):
            halp[namespace] = {}

            for member_schema in schema[namespace].values():
                member_full_name = member_schema['full_name']

                topic = halp[namespace].setdefault(member_full_name, {})
                topic['name'] = member_schema['name']
                if 'doc' in member_schema:
                    topic['summary'] = (
                        member_schema['doc'].split('\n\n', 1)[0].strip())
                if 'topic_topic' in member_schema:
                    topic['topic_topic'] = member_schema['topic_topic']
                if 'exclude' in member_schema:
                    topic['exclude'] = member_schema['exclude']

        return halp

    def _write_schema(self, fingerprint):
        try:
            os.makedirs(self._DIR)
        except EnvironmentError as e:
            if e.errno != errno.EEXIST:
                raise

        with tempfile.NamedTemporaryFile('wb', prefix=fingerprint,
                                         dir=self._DIR, delete=False) as f:
            try:
                self._write_schema_data(f)
                ipautil.flush_sync(f)
                f.close()
            except Exception:
                os.unlink(f.name)
                raise
            else:
                os.rename(f.name, os.path.join(self._DIR, fingerprint))

    def _write_schema_data(self, fileobj):
        with zipfile.ZipFile(fileobj, 'w', zipfile.ZIP_DEFLATED) as schema:
            for key, value in self._dict.items():
                if key in self.namespaces:
                    ns = value
                    for member in ns:
                        path = '{}/{}'.format(key, member)
                        s = json.dumps(ns[member], default=json_default)
                        schema.writestr(path, s.encode('utf-8'))
                else:
                    schema.writestr(key, json.dumps(value).encode('utf-8'))

            schema.writestr(
                '_help',
                json.dumps(self._help, default=json_default).encode('utf-8')
            )

    def read_namespace_member(self, namespace, member):
        value = self._dict[namespace][member]

        if isinstance(value, bytes):
            value = json.loads(value.decode('utf-8'))
            self._dict[namespace][member] = value

        return value

    def iter_namespace(self, namespace):
        return iter(self._dict[namespace])

    def get_help(self, namespace, member):
        if isinstance(self._help, bytes):
            self._help = json.loads(
                self._help.decode('utf-8')  # pylint: disable=no-member
            )

        return self._help[namespace][member]


def get_package(server_info, client):
    NO_FINGERPRINT = object()

    fingerprint = NO_FINGERPRINT
    if server_info.is_valid():
        fingerprint = server_info.get('fingerprint', fingerprint)

    if fingerprint is not None:
        try:
            try:
                if fingerprint is NO_FINGERPRINT:
                    schema = Schema(client)
                else:
                    schema = Schema(client, fingerprint)
            except SchemaUpToDate as e:
                schema = Schema(client, e.fingerprint)
        except NotAvailable:
            fingerprint = None
            ttl = None
        except SchemaUpToDate as e:
            fingerprint = e.fingerprint
            ttl = e.ttl
        else:
            fingerprint = schema.fingerprint
            ttl = schema.ttl

        server_info['fingerprint'] = fingerprint
        server_info.update_validity(ttl)

    if fingerprint is None:
        raise NotAvailable()

    fingerprint = str(fingerprint)
    package_name = '{}${}'.format(__name__, fingerprint)
    package_dir = '{}${}'.format(os.path.splitext(__file__)[0], fingerprint)

    try:
        return sys.modules[package_name]
    except KeyError:
        pass

    package = types.ModuleType(package_name)
    package.__file__ = os.path.join(package_dir, '__init__.py')
    package.modules = ['plugins']
    sys.modules[package_name] = package

    module_name = '.'.join((package_name, 'plugins'))
    module = types.ModuleType(module_name)
    module.__file__ = os.path.join(package_dir, 'plugins.py')
    module.register = plugable.Registry()
    for plugin_cls in (_SchemaCommandPlugin, _SchemaObjectPlugin):
        for full_name in schema[plugin_cls.schema_key]:
            plugin = plugin_cls(schema, str(full_name))
            plugin = module.register()(plugin)  # pylint: disable=no-member
    sys.modules[module_name] = module

    for full_name, topic in six.iteritems(schema['topics']):
        name = str(topic['name'])
        module_name = '.'.join((package_name, name))
        try:
            module = sys.modules[module_name]
        except KeyError:
            module = sys.modules[module_name] = types.ModuleType(module_name)
            module.__file__ = os.path.join(package_dir, '{}.py'.format(name))
        module.__doc__ = topic.get('doc')
        if 'topic_topic' in topic:
            s = topic['topic_topic']
            if isinstance(s, bytes):
                s = s.decode('utf-8')
            module.topic = str(s).partition('/')[0]
        else:
            module.topic = None

    return package



# Authors:
#   Petr Vobornik <pvoborni@redhat.com>
#
# Copyright (C) 2013  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Host tests
"""

import uuid
from random import randint

from ipatests.test_webui.crypto_utils import generate_certificate, generate_csr
from ipatests.test_webui.ui_driver import UI_driver
from ipatests.test_webui.ui_driver import screenshot
import ipatests.test_webui.data_hostgroup as hostgroup
import ipatests.test_webui.data_netgroup as netgroup
import ipatests.test_webui.data_hbac as hbac
import ipatests.test_webui.test_rbac as rbac
import ipatests.test_webui.data_sudo as sudo
import ipatests.test_webui.data_host as host
import pytest

try:
    from selenium.webdriver.common.by import By
    from selenium.webdriver.common.keys import Keys
    from selenium.webdriver.common.action_chains import ActionChains
except ImportError:
    NO_SELENIUM = True

ENTITY = 'host'


@pytest.mark.tier1
class host_tasks(UI_driver):

    def setup(self, *args, **kwargs):
        super(host_tasks, self).setup(*args, **kwargs)
        self.prep_data()
        self.prep_data2()
        self.prep_data3()
        self.prep_data4()

    def prep_data(self):
        host = self.rand_host()
        domain = self.config.get('ipa_domain')
        ip = self.get_ip()
        self.data = self.get_data(host, domain, ip)
        self.pkey = self.data['pkey']
        return self.data

    def prep_data2(self):
        host = self.rand_host()
        domain = self.config.get('ipa_domain')
        self.data2 = self.get_data(host, domain)
        self.pkey2 = self.data2['pkey']
        return self.data2

    def prep_data3(self):
        host = self.rand_host()
        domain = self.config.get('ipa_domain')
        self.data3 = self.get_data(host, domain)
        self.pkey3 = self.data3['pkey']
        return self.data3

    def prep_data4(self):
        host = self.rand_host()
        domain = self.config.get('ipa_domain')
        self.data4 = self.get_data(host, domain)
        self.pkey4 = self.data4['pkey']
        return self.data4

    def get_data(self, host, domain, ip=None):
        if self.has_dns():
            add_data = [
                ('textbox', 'hostname', host),
                ('combobox', 'dnszone', domain+'.'),
            ]
            if ip:
                add_data.append(('textbox', 'ip_address', ip))
            add_data.append(('checkbox', 'force', None))
            del_data = [
                ('checkbox', 'updatedns', None)
            ]
        else:
            add_data = [
                ('textbox', 'fqdn', '%s.%s' % (host, domain)),
                ('checkbox', 'force', None),
            ]
            del_data = None

        data = {
            'pkey': '%s.%s' % (host, domain),
            'add': add_data,
            'mod': [
                ('textarea', 'description', 'Desc'),
            ],
            'del': del_data,
        }

        return data

    def get_ip(self):
        """
        Get next IP
        """
        ip = self.config.get('ipa_ip')
        if not ip:
            self.skip('FreeIPA Server IP address not configured')

        while True:
            new_ip = '10.{}.{}.{}'.format(
                randint(0, 255),
                randint(0, 255),
                randint(1, 254)
            )
            if new_ip != ip:
                break
        return new_ip

    @staticmethod
    def rand_host():
        return 'host-{}'.format(uuid.uuid4().hex[:8])

    def load_file(self, path):
        """
        Load file helper mainly for CSR load_file
        """

        with open(path, 'r') as file_d:
            content = file_d.read()
        return content


@pytest.mark.tier1
class test_host(host_tasks):

    @screenshot
    def test_crud(self):
        """
        Basic CRUD: host
        """
        self.init_app()
        self.basic_crud(ENTITY, self.data)

    @screenshot
    def test_certificates(self):
        """
        Test host certificate actions
        """

        if not self.has_ca():
            self.skip('CA is not configured')

        self.init_app()

        cert_widget_sel = "div.certificate-widget"

        self.add_record(ENTITY, self.data)
        self.navigate_to_record(self.pkey)

        # cert request
        csr = generate_csr(self.pkey)
        self.action_list_action('request_cert', confirm=False)
        self.assert_dialog()
        self.fill_text("textarea[name='csr']", csr)
        self.dialog_button_click('issue')
        self.wait_for_request(n=2, d=3)
        self.assert_visible(cert_widget_sel)

        widget = self.find(cert_widget_sel, By.CSS_SELECTOR)

        # cert view
        self.action_list_action('view', confirm=False,
                                parents_css_sel=cert_widget_sel)
        self.assert_dialog()
        self.dialog_button_click('close')

        # cert get
        self.action_list_action('get', confirm=False,
                                parents_css_sel=cert_widget_sel)
        self.assert_dialog()
        # check that the textarea is not empty
        self.assert_empty_value('textarea.certificate', negative=True)
        self.dialog_button_click('close')

        # cert download - we can only try to click the download action
        self.action_list_action('download', confirm=False,
                                parents_css_sel=cert_widget_sel)

        # check that revoke action is enabled
        self.assert_action_list_action('revoke',
                                       parents_css_sel=cert_widget_sel,
                                       facet_actions=False)

        # check that remove_hold action is not enabled
        self.assert_action_list_action('remove_hold', enabled=False,
                                       parents_css_sel=cert_widget_sel,
                                       facet_actions=False)

        # cert revoke
        self.action_list_action('revoke', confirm=False,
                                parents_css_sel=cert_widget_sel)
        self.wait()
        self.select('select', '6')
        self.dialog_button_click('ok')
        self.wait_while_working(widget)

        self.assert_visible(cert_widget_sel + " div.watermark")

        # check that revoke action is not enabled
        self.assert_action_list_action('revoke', enabled=False,
                                       parents_css_sel=cert_widget_sel,
                                       facet_actions=False)

        # check that remove_hold action is enabled
        self.assert_action_list_action('remove_hold',
                                       parents_css_sel=cert_widget_sel,
                                       facet_actions=False)

        # cert remove hold
        self.action_list_action('remove_hold', confirm=False,
                                parents_css_sel=cert_widget_sel)
        self.wait()
        self.dialog_button_click('ok')
        self.wait_while_working(widget)

        # check that revoke action is enabled
        self.assert_action_list_action('revoke',
                                       parents_css_sel=cert_widget_sel,
                                       facet_actions=False)

        # check that remove_hold action is not enabled
        self.assert_action_list_action('remove_hold', enabled=False,
                                       parents_css_sel=cert_widget_sel,
                                       facet_actions=False)

        # cleanup
        self.navigate_to_entity(ENTITY, 'search')
        self.delete_record(self.pkey, self.data.get('del'))

    @screenshot
    def test_arbitrary_certificates(self):
        """
        Test managing host arbitrary certificate.
        """
        self.init_app()
        self.add_record(ENTITY, self.data)

        self.navigate_to_record(self.pkey)

        # check whether certificate section is present
        self.assert_visible("div[name='certificate']")

        # add certificate
        self.button_click('add', parents_css_sel="div[name='certificate']")
        self.assert_dialog('cert-add-dialog')

        cert = generate_certificate(self.pkey)
        self.fill_textarea('new_cert', cert)
        self.dialog_button_click('ok')

        self.assert_visible("div.certificate-widget")

        # cert view
        self.action_list_action('view', confirm=False,
                                parents_css_sel="div.certificate-widget")
        self.assert_dialog()
        self.dialog_button_click('close')

        # cert get
        self.action_list_action('get', confirm=False,
                                parents_css_sel="div.certificate-widget")
        self.assert_dialog()

        # check that the textarea is not empty
        self.assert_empty_value('textarea.certificate', negative=True)
        self.dialog_button_click('close')

        # cert download - we can only try to click the download action
        self.action_list_action('download', confirm=False,
                                parents_css_sel="div.certificate-widget")

        # check that revoke action is not enabled
        self.assert_action_list_action(
            'revoke', enabled=False,
            parents_css_sel="div.certificate-widget",
            facet_actions=False)

        # check that remove_hold action is not enabled
        self.assert_action_list_action(
            'remove_hold', enabled=False,
            parents_css_sel="div.certificate-widget",
            facet_actions=False)

        # cleanup
        self.navigate_to_entity(ENTITY, 'search')
        self.delete_record(self.pkey, self.data.get('del'))

    @screenshot
    def test_ca_less(self):
        """
        Test host certificate actions in CA-less install
        http://www.freeipa.org/page/V3/CA-less_install
        """
        if self.has_ca():
            self.skip('CA is installed')

        self.init_app()
        self.add_record(ENTITY, self.data)
        self.navigate_to_record(self.pkey)

        self.assert_action_list_action('request_cert', visible=False)

        self.navigate_by_breadcrumb('Hosts')
        self.delete_record(self.pkey, self.data.get('del'))

    @screenshot
    def test_kerberos_flags(self):
        """
        Test Kerberos flags
        http://www.freeipa.org/page/V3/Kerberos_Flags
        """
        name = 'ipakrbokasdelegate'
        mod = {'mod': [('checkbox', name, None)]}
        checked = ['checked']

        self.init_app()
        self.add_record(ENTITY, self.data)
        self.navigate_to_record(self.pkey)

        if self.get_field_checked(name) == checked:
            self.mod_record(ENTITY, mod)  # uncheck

        self.mod_record(ENTITY, mod)
        self.validate_fields([('checkbox', name, checked)])
        self.mod_record(ENTITY, mod)
        self.validate_fields([('checkbox', name, [])])
        self.close_notifications()
        self.delete(ENTITY, [self.data])

    @screenshot
    def test_associations(self):
        """
        Host direct associations
        """

        self.init_app()

        # prepare
        # -------
        self.add_record(ENTITY, self.data)
        self.add_record(ENTITY, self.data2, navigate=False)
        self.add_record(hostgroup.ENTITY, hostgroup.DATA)
        self.add_record(netgroup.ENTITY, netgroup.DATA)
        self.add_record(rbac.ROLE_ENTITY, rbac.ROLE_DATA)
        self.add_record(hbac.RULE_ENTITY, hbac.RULE_DATA)
        self.add_record(sudo.RULE_ENTITY, sudo.RULE_DATA)

        # add & remove associations
        # -------------------------
        self.navigate_to_entity(ENTITY)
        self.navigate_to_record(self.pkey)

        self.add_associations([hostgroup.PKEY], facet='memberof_hostgroup', delete=True)
        self.add_associations([netgroup.PKEY], facet='memberof_netgroup', delete=True)
        self.add_associations([rbac.ROLE_PKEY], facet='memberof_role', delete=True)
        self.add_associations([hbac.RULE_PKEY], facet='memberof_hbacrule', delete=True)
        self.add_associations([sudo.RULE_PKEY], facet='memberof_sudorule', delete=True)
        self.add_associations([self.pkey2], facet='managedby_host', delete=True)

        # cleanup
        # -------
        self.delete(ENTITY, [self.data, self.data2])
        self.delete(hostgroup.ENTITY, [hostgroup.DATA])
        self.delete(netgroup.ENTITY, [netgroup.DATA])
        self.delete(rbac.ROLE_ENTITY, [rbac.ROLE_DATA])
        self.delete(hbac.RULE_ENTITY, [hbac.RULE_DATA])
        self.delete(sudo.RULE_ENTITY, [sudo.RULE_DATA])

    @screenshot
    def test_indirect_associations(self):
        """
        Host indirect associations
        """
        self.init_app()

        # add
        # ---
        self.add_record(ENTITY, self.data)

        self.add_record(hostgroup.ENTITY, hostgroup.DATA)
        self.navigate_to_record(hostgroup.PKEY)
        self.add_associations([self.pkey])

        self.add_record(hostgroup.ENTITY, hostgroup.DATA2)
        self.navigate_to_record(hostgroup.PKEY2)
        self.switch_to_facet('member_hostgroup')
        self.add_associations([hostgroup.PKEY])

        self.add_record(netgroup.ENTITY, netgroup.DATA)
        self.navigate_to_record(netgroup.PKEY)
        self.add_table_associations('memberhost_hostgroup', [hostgroup.PKEY2])

        self.add_record(rbac.ROLE_ENTITY, rbac.ROLE_DATA)
        self.navigate_to_record(rbac.ROLE_PKEY)
        self.switch_to_facet('member_hostgroup')
        self.add_associations([hostgroup.PKEY2])

        self.add_record(hbac.RULE_ENTITY, hbac.RULE_DATA)
        self.navigate_to_record(hbac.RULE_PKEY)
        self.add_table_associations('memberhost_hostgroup', [hostgroup.PKEY2])

        self.add_record(sudo.RULE_ENTITY, sudo.RULE_DATA)
        self.navigate_to_record(sudo.RULE_PKEY)
        self.add_table_associations('memberhost_hostgroup', [hostgroup.PKEY2])

        # check indirect associations
        # ---------------------------
        self.navigate_to_entity(ENTITY, 'search')
        self.navigate_to_record(self.pkey)

        self.assert_indirect_record(hostgroup.PKEY2, ENTITY, 'memberof_hostgroup')
        self.assert_indirect_record(netgroup.PKEY, ENTITY, 'memberof_netgroup')
        self.assert_indirect_record(rbac.ROLE_PKEY, ENTITY, 'memberof_role')
        self.assert_indirect_record(hbac.RULE_PKEY, ENTITY, 'memberof_hbacrule')
        self.assert_indirect_record(sudo.RULE_PKEY, ENTITY, 'memberof_sudorule')

        # cleanup
        # -------
        self.delete(ENTITY, [self.data])
        self.delete(hostgroup.ENTITY, [hostgroup.DATA, hostgroup.DATA2])
        self.delete(netgroup.ENTITY, [netgroup.DATA])
        self.delete(rbac.ROLE_ENTITY, [rbac.ROLE_DATA])
        self.delete(hbac.RULE_ENTITY, [hbac.RULE_DATA])
        self.delete(sudo.RULE_ENTITY, [sudo.RULE_DATA])

    @screenshot
    def test_buttons(self):
        """Test buttons"""
        self.init_app()
        self.navigate_to_entity(ENTITY)
        # add with enter key
        self.navigate_to_entity(ENTITY)
        self.button_click('add')
        self.fill_textbox("hostname", self.pkey)
        self.check_option('force')
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait_for_request(n=3)
        self.assert_record(self.pkey)

        # add and another
        self.add_record(ENTITY, [self.data2, self.data3])

        # add and edit record
        self.add_record(ENTITY, self.data4, dialog_btn='add_and_edit')
        self.assert_facet(ENTITY, facet="details")

        # cancel managedby
        self.add_associations([self.pkey3], facet='managedby_host',
                              confirm_btn="cancel")
        self.wait()
        self.select_record(self.pkey4, table_name='fqdn')
        self.button_click('remove')
        self.dialog_button_click('cancel')
        self.assert_record(self.pkey4)

        # add duplicate
        self.add_record(ENTITY, self.data2, negative=True, pre_delete=False)
        dialog_info = self.get_dialog_info()
        expected_msg = 'host with name "' + self.data2['pkey'] + \
                       '" already exist'
        if expected_msg in dialog_info['text']:
            self.dialog_button_click('cancel')
            self.dialog_button_click('cancel')
        else:
            assert False, "Duplicate dialog missing or have wrong text."

        # duplicate with pressed keys
        self.add_record(ENTITY, self.data2, negative=True, pre_delete=False)
        self.wait()
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.TAB).perform()
        self.wait()
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait(2)
        self.dialog_button_click('cancel')
        self.assert_no_dialog()

        # remove multiple
        self.navigate_to_entity(ENTITY)
        self.select_multiple_records([self.data2, self.data3, self.data4])
        self.facet_button_click('remove')
        self.wait()
        self.check_option('updatedns')
        self.dialog_button_click('ok')
        self.assert_notification()
        self.close_notifications()

        # remove without updatedns
        self.select_record(self.pkey)
        self.facet_button_click('remove')
        self.dialog_button_click('ok')
        self.assert_notification()
        self.close_notifications()

    @screenshot
    def test_negative_add_input(self):
        """ Test field validations for adding """
        self.init_app()

        # wrong hostname input
        hosts_tests = [host.hostname_tilde,
                       host.hostname_trailing_space,
                       host.hostname_dash]
        for hostname_test in hosts_tests:
            self.add_record(ENTITY, hostname_test, negative=True)
            dialog_info = self.get_dialog_info()
            if host.BAD_HOSTNAME_MSG in dialog_info['text']:
                self.dialog_button_click('cancel')
                self.dialog_button_click('cancel')

        # leading space in hostname
        self.add_record(ENTITY, host.hostname_leading_space, negative=True)
        dialog_info = self.get_dialog_info()
        if host.BAS_HOSTNAME_SPACE_MSG in dialog_info['text']:
            self.dialog_button_click('cancel')
            self.dialog_button_click('cancel')

        # empty hostname
        self.add_record(ENTITY, host.empty_hostname, negative=True)
        self.assert_field_validation_required(field='hostname')
        self.dialog_button_click('cancel')

        # empty domain
        self.add_record(ENTITY, host.empty_domain, negative=True)
        self.assert_field_validation_required(field='dnszone')
        self.dialog_button_click('cancel')

        # Wrong IP input
        ip_tests = [host.ip_alpha, host.ip_many_oct,
                    host.ip_bad_oct, host.ip_special_char]
        for ip_test in ip_tests:
            self.add_record(ENTITY, ip_test, negative=True)
            self.assert_field_validation(host.BAD_IP_MSG, field='ip_address')
            self.dialog_button_click('cancel')

    @screenshot
    def test_details_input(self):
        """ Test text fields in details page """
        self.init_app()

        self.add_record(ENTITY, self.data2)
        self.navigate_to_record(self.data2['pkey'], entity=ENTITY)

        # modify
        modify_tests = [host.mod_desc, host.mod_locality,
                        host.mod_location, host.mod_platform,
                        host.mod_os]
        for mod_test in modify_tests:
            self.fill_fields(mod_test)
            self.button_click('save')
            self.assert_notification()
            self.close_notifications()

        self.fill_fields(host.mod_desc_m)
        self.click_undo_button('description')

        self.delete(ENTITY, [self.data2])

        # otp set_otp
        otp_tests = [host.otp_alpha, host.otp_num, host.otp_alphanum,
                     host.otp_special, host.otp_mixed]
        for otp_test in otp_tests:
            self.add_record(ENTITY, self.data2)
            self.close_notifications()
            self.navigate_to_record(self.data2['pkey'], entity=ENTITY)
            self.action_list_action('set_otp', confirm=False)
            self.fill_fields(otp_test)
            self.dialog_button_click('confirm')
            self.assert_notification()
            self.close_notifications()
            self.delete(ENTITY, [self.data2])
            self.close_notifications()

        # otp cancel and reset
        self.add_record(ENTITY, self.data2)
        self.navigate_to_record(self.data2['pkey'], entity=ENTITY)
        self.action_list_action('set_otp', confirm=False)
        self.assert_dialog()
        self.dialog_button_click('cancel')
        self.assert_no_dialog()
        self.navigate_to_record(self.data2['pkey'], entity=ENTITY)
        self.action_list_action('set_otp', confirm=False)
        self.fill_fields(host.otp_alpha)
        self.dialog_button_click('confirm')
        self.assert_notification()
        self.close_notifications()
        self.action_list_action('reset_otp', confirm=False)
        self.assert_dialog()
        self.dialog_button_click('cancel')
        self.assert_no_dialog()

        # cleanup
        self.delete(ENTITY, [self.data2])

    @screenshot
    def test_sshkey(self):
        """ Test ssh keys """
        self.init_app()
        self.add_record(ENTITY, self.data2)
        self.close_notifications()
        # add dsa key
        self.add_sshkey_to_record(host.ssh_dsa, self.data2['pkey'],
                                  entity=ENTITY, navigate=True)
        self.assert_notification()
        self.close_notifications()

        # delete ssh key
        self.delete_record_sshkeys(self.data2['pkey'],
                                   entity=ENTITY, navigate=True)

        # add rsa key
        self.add_sshkey_to_record(host.ssh_rsa, self.data2['pkey'],
                                  entity=ENTITY, navigate=True)
        self.assert_notification()
        self.close_notifications()

        # negative ssh key input
        neg_key_tests = [host.ssh_empty, host.ssh_rsa]
        for key in neg_key_tests:
            self.add_sshkey_to_record(key, self.data2['pkey'],
                                      entity=ENTITY, navigate=True)
            self.assert_dialog()
            dialog_info = self.get_dialog_info()
            if host.ssh_nomod_error in dialog_info['text']:
                self.dialog_button_click('cancel')

        # invalid ssh key
        self.add_sshkey_to_record(host.ssh_invalid, self.data2['pkey'],
                                  entity=ENTITY, navigate=True)
        self.assert_dialog()
        dialog_info = self.get_dialog_info()
        if host.ssh_invalid_error in dialog_info['text']:
            self.dialog_button_click('cancel')

        # undo all and delete ssh keys
        self.undo_ssh_keys(btn_name='undo_all')
        self.delete_record_sshkeys(self.data2['pkey'],
                                   entity=ENTITY)

        # undo
        self.add_sshkey_to_record(host.ssh_rsa, self.data2['pkey'],
                                  entity=ENTITY, navigate=True, save=False)
        self.undo_ssh_keys()

        # refresh
        self.add_sshkey_to_record(host.ssh_rsa, self.data2['pkey'],
                                  entity=ENTITY, navigate=True, save=False)
        self.facet_button_click('refresh')
        self.assert_num_ssh_keys(0)

        # revert
        self.add_sshkey_to_record(host.ssh_rsa, self.data2['pkey'],
                                  entity=ENTITY, navigate=True, save=False)
        self.facet_button_click('revert')
        self.assert_num_ssh_keys(0)

        # cleanup
        self.delete(ENTITY, [self.data2])

    @screenshot
    def test_negative_cert(self):
        """ Test negative CSR """
        self.init_app()
        self.add_record(ENTITY, self.data2)
        self.close_notifications()
        self.navigate_to_record(self.data2['pkey'], entity=ENTITY)

        # emtpy CSR
        csr_add = 'div[name="certificate"] button[name="add"]'
        csr_add_btn = self.find(csr_add, By.CSS_SELECTOR, strict=True)
        csr_add_btn.click()
        self.wait()
        self.dialog_button_click('ok')
        self.assert_field_validation_required()
        self.dialog_button_click('cancel')

        # invalid CSR
        csr_add = 'div[name="certificate"] button[name="add"]'
        csr_add_btn = self.find(csr_add, By.CSS_SELECTOR, strict=True)
        csr_add_btn.click()
        self.wait()
        self.fill_textarea('new_cert', host.csr_invalid)
        self.dialog_button_click('ok')
        dialog_info = self.get_dialog_info()
        self.wait()
        if host.csr_invalid_error in dialog_info['text']:
            self.dialog_button_click('cancel')
            self.dialog_button_click('cancel')

        # other hostname CSR
        self.action_list_action('request_cert', confirm=False)
        self.assert_dialog()
        self.fill_text("textarea[name='csr']", host.csr_other_host)
        self.dialog_button_click('issue')
        dialog_info = self.get_dialog_info()
        if host.csr_other_host_error in dialog_info['text']:
            self.dialog_button_click('cancel')
            self.dialog_button_click('cancel')

        # cleanup
        self.delete(ENTITY, [self.data2])

    @screenshot
    def test_keytab(self):
        """ Test keytab """
        self.init_app()
        self.add_record(ENTITY, self.data2)
        # provision keytab
        kt_tmp = '/tmp/test.keytab'
        hostname = self.data2['pkey']
        realm = self.config.get('ipa_realm')
        principal = 'host/{}@{}'.format(hostname, realm)
        self.run_cmd_on_ui_host('/usr/sbin/ipa-getkeytab '
                                '-p {} '
                                '-k {} '.format(principal, kt_tmp))
        self.navigate_to_record(hostname, entity=ENTITY)
        self.wait(3)
        enroll = 'div[name="has_keytab"] label[name="present"]'
        self.assert_text(enroll, host.krb_enrolled)
        # test cancel button exist
        self.action_list_action('unprovision', confirm=False)
        self.dialog_button_click('cancel')
        # unprovision keytab
        self.action_list_action('unprovision', confirm=False)
        self.dialog_button_click('unprovision')
        self.wait_for_request(n=4)
        self.facet_button_click('refresh')
        enroll = 'div[name="has_keytab"] label[name="missing"]'
        self.assert_text(enroll, host.krb_not_enrolled)
        # cleanup
        self.delete(ENTITY, [self.data2])

    def test_search(self):
        self.init_app()
        self.navigate_to_entity(ENTITY)
        self.add_record(ENTITY, [self.data2, self.data3])
        # positive search filter
        self.fill_search_filter(self.pkey2)
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait(0.5)
        self.assert_record(self.pkey2)
        # negative search filter
        self.fill_search_filter(self.pkey3)
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait(0.5)
        self.assert_record(self.pkey4, negative=True)
        # cleanup
        self.fill_search_filter('')
        actions = ActionChains(self.driver)
        actions.send_keys(Keys.ENTER).perform()
        self.wait(0.5)
        self.delete_record([self.pkey2, self.pkey3])



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
Automount

Stores automount(8) configuration for autofs(8) in IPA.

The base of an automount configuration is the configuration file auto.master.
This is also the base location in IPA. Multiple auto.master configurations
can be stored in separate locations. A location is implementation-specific
with the default being a location named 'default'. For example, you can have
locations by geographic region, by floor, by type, etc.

Automount has three basic object types: locations, maps and keys.

A location defines a set of maps anchored in auto.master. This allows you
to store multiple automount configurations. A location in itself isn't
very interesting, it is just a point to start a new automount map.

A map is roughly equivalent to a discrete automount file and provides
storage for keys.

A key is a mount point associated with a map.

When a new location is created, two maps are automatically created for
it: auto.master and auto.direct. auto.master is the root map for all
automount maps for the location. auto.direct is the default map for
direct mounts and is mounted on /-.

An automount map may contain a submount key. This key defines a mount
location within the map that references another map. This can be done
either using automountmap-add-indirect --parentmap or manually
with automountkey-add and setting info to "-type=autofs :<mapname>".

EXAMPLES:

Locations:

  Create a named location, "Baltimore":
    ipa automountlocation-add baltimore

  Display the new location:
    ipa automountlocation-show baltimore

  Find available locations:
    ipa automountlocation-find

  Remove a named automount location:
    ipa automountlocation-del baltimore

  Show what the automount maps would look like if they were in the filesystem:
    ipa automountlocation-tofiles baltimore

  Import an existing configuration into a location:
    ipa automountlocation-import baltimore /etc/auto.master

    The import will fail if any duplicate entries are found. For
    continuous operation where errors are ignored, use the --continue
    option.

Maps:

  Create a new map, "auto.share":
    ipa automountmap-add baltimore auto.share

  Display the new map:
    ipa automountmap-show baltimore auto.share

  Find maps in the location baltimore:
    ipa automountmap-find baltimore

  Create an indirect map with auto.share as a submount:
    ipa automountmap-add-indirect baltimore --parentmap=auto.share --mount=sub auto.man

    This is equivalent to:

    ipa automountmap-add-indirect baltimore --mount=/man auto.man
    ipa automountkey-add baltimore auto.man --key=sub --info="-fstype=autofs ldap:auto.share"

  Remove the auto.share map:
    ipa automountmap-del baltimore auto.share

Keys:

  Create a new key for the auto.share map in location baltimore. This ties
  the map we previously created to auto.master:
    ipa automountkey-add baltimore auto.master --key=/share --info=auto.share

  Create a new key for our auto.share map, an NFS mount for man pages:
    ipa automountkey-add baltimore auto.share --key=man --info="-ro,soft,rsize=8192,wsize=8192 ipa.example.com:/shared/man"

  Find all keys for the auto.share map:
    ipa automountkey-find baltimore auto.share

  Find all direct automount keys:
    ipa automountkey-find baltimore --key=/-

  Remove the man key from the auto.share map:
    ipa automountkey-del baltimore auto.share --key=man
""")

register = Registry()


@register()
class automountkey(Object):
    takes_params = (
        parameters.Str(
            'automountkey',
            label=_(u'Key'),
            doc=_(u'Automount key name.'),
        ),
        parameters.Str(
            'automountinformation',
            label=_(u'Mount information'),
        ),
        parameters.Str(
            'description',
            required=False,
            primary_key=True,
            label=_(u'description'),
            exclude=('webui', 'cli'),
        ),
    )


@register()
class automountlocation(Object):
    takes_params = (
        parameters.Str(
            'cn',
            primary_key=True,
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
    )


@register()
class automountmap(Object):
    takes_params = (
        parameters.Str(
            'automountmapname',
            primary_key=True,
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
        ),
    )


@register()
class automountkey_add(Method):
    __doc__ = _("Create a new automount key.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapautomountmapname',
            cli_name='automountmap',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Str(
            'automountkey',
            cli_name='key',
            label=_(u'Key'),
            doc=_(u'Automount key name.'),
        ),
        parameters.Str(
            'automountinformation',
            cli_name='info',
            label=_(u'Mount information'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountkey_del(Method):
    __doc__ = _("Delete an automount key.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapautomountmapname',
            cli_name='automountmap',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'automountkey',
            cli_name='key',
            label=_(u'Key'),
            doc=_(u'Automount key name.'),
        ),
        parameters.Str(
            'automountinformation',
            required=False,
            cli_name='info',
            label=_(u'Mount information'),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class automountkey_find(Method):
    __doc__ = _("Search for an automount key.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapautomountmapname',
            cli_name='automountmap',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'automountkey',
            required=False,
            cli_name='key',
            label=_(u'Key'),
            doc=_(u'Automount key name.'),
        ),
        parameters.Str(
            'automountinformation',
            required=False,
            cli_name='info',
            label=_(u'Mount information'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class automountkey_mod(Method):
    __doc__ = _("Modify an automount key.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapautomountmapname',
            cli_name='automountmap',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Str(
            'automountkey',
            cli_name='key',
            label=_(u'Key'),
            doc=_(u'Automount key name.'),
        ),
        parameters.Str(
            'automountinformation',
            required=False,
            cli_name='info',
            label=_(u'Mount information'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'newautomountinformation',
            required=False,
            cli_name='newinfo',
            label=_(u'New mount information'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'rename',
            required=False,
            label=_(u'Rename'),
            doc=_(u'Rename the automount key object'),
            exclude=('webui',),
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountkey_show(Method):
    __doc__ = _("Display an automount key.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapautomountmapname',
            cli_name='automountmap',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Str(
            'automountkey',
            cli_name='key',
            label=_(u'Key'),
            doc=_(u'Automount key name.'),
        ),
        parameters.Str(
            'automountinformation',
            required=False,
            cli_name='info',
            label=_(u'Mount information'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountlocation_add(Method):
    __doc__ = _("Create a new automount location.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='location',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
    )
    takes_options = (
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountlocation_del(Method):
    __doc__ = _("Delete an automount location.")

    takes_args = (
        parameters.Str(
            'cn',
            multivalue=True,
            cli_name='location',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class automountlocation_find(Method):
    __doc__ = _("Search for an automount location.")

    takes_args = (
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'cn',
            required=False,
            cli_name='location',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("location")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class automountlocation_show(Method):
    __doc__ = _("Display an automount location.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='location',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountlocation_tofiles(Method):
    __doc__ = _("Generate automount files for a specific location.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='location',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
    )
    takes_options = (
    )
    has_output = (
        output.Output(
            'result',
        ),
    )


@register()
class automountmap_add(Method):
    __doc__ = _("Create a new automount map.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapname',
            cli_name='map',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountmap_add_indirect(Method):
    __doc__ = _("Create a new indirect mount point.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapname',
            cli_name='map',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'key',
            cli_name='mount',
            label=_(u'Mount point'),
        ),
        parameters.Str(
            'parentmap',
            required=False,
            label=_(u'Parent map'),
            doc=_(u'Name of parent automount map (default: auto.master).'),
            default=u'auto.master',
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountmap_del(Method):
    __doc__ = _("Delete an automount map.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapname',
            multivalue=True,
            cli_name='map',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class automountmap_find(Method):
    __doc__ = _("Search for an automount map.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'automountmapname',
            required=False,
            cli_name='map',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("map")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class automountmap_mod(Method):
    __doc__ = _("Modify an automount map.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapname',
            cli_name='map',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class automountmap_show(Method):
    __doc__ = _("Display an automount map.")

    takes_args = (
        parameters.Str(
            'automountlocationcn',
            cli_name='automountlocation',
            label=_(u'Location'),
            doc=_(u'Automount location name.'),
        ),
        parameters.Str(
            'automountmapname',
            cli_name='map',
            label=_(u'Map'),
            doc=_(u'Automount map name.'),
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )



#
# Copyright (C) 2019  IPA Project Contributors, see COPYING for license
#
"""Common helpers for handlers
"""
import argparse
import base64
import json
import shutil
import tempfile


def default_json(obj):
    """JSON encoder default handler
    """
    if isinstance(obj, (bytes, bytearray)):
        return base64.b64encode(obj).decode('ascii')
    raise TypeError(
        "Object of type {} is not JSON serializable".format(type(obj))
    )


def json_dump(data, exportfile):
    """Dump JSON to file
    """
    json.dump(
        data,
        exportfile,
        default=default_json,
        separators=(',', ':'),
        sort_keys=True
    )


def mkparser(supports_import=True, **kwargs):
    """Create default parser for handler with export / import args

    All commands support export to file or stdout. Most commands can also
    import from a file or stdin. Export and import are mutually exclusive
    options.
    """
    parser = argparse.ArgumentParser(**kwargs)

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--export',
        help='JSON export file ("-" for stdout)',
        dest='exportfile',
        type=argparse.FileType('w')
    )
    if supports_import:
        group.add_argument(
            '--import',
            help='JSON import file ("-" for stdin)',
            dest='importfile',
            type=argparse.FileType('r')
        )

    return parser


def main(parser, export_func, import_func=None, **kwargs):
    """Common main function for handlers
    """
    args = parser.parse_args()
    if args.exportfile is not None:
        func = export_func
    else:
        func = import_func

    tmpdir = tempfile.mkdtemp()
    try:
        func(args, tmpdir, **kwargs)
    finally:
        shutil.rmtree(tmpdir)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

# pylint: disable=unused-import
import six

from . import Command, Method, Object
from ipalib import api, parameters, output
from ipalib.parameters import DefaultFrom
from ipalib.plugable import Registry
from ipalib.text import _
from ipapython.dn import DN
from ipapython.dnsutil import DNSName

if six.PY3:
    unicode = str

__doc__ = _("""
HBAC Services

The PAM services that HBAC can control access to. The name used here
must match the service name that PAM is evaluating.

EXAMPLES:

 Add a new HBAC service:
   ipa hbacsvc-add tftp

 Modify an existing HBAC service:
   ipa hbacsvc-mod --desc="TFTP service" tftp

 Search for HBAC services. This example will return two results, the FTP
 service and the newly-added tftp service:
   ipa hbacsvc-find ftp

 Delete an HBAC service:
   ipa hbacsvc-del tftp
""")

register = Registry()


@register()
class hbacsvc(Object):
    takes_params = (
        parameters.Str(
            'cn',
            primary_key=True,
            label=_(u'Service name'),
            doc=_(u'HBAC service'),
        ),
        parameters.Str(
            'description',
            required=False,
            label=_(u'Description'),
            doc=_(u'HBAC service description'),
        ),
        parameters.Str(
            'memberof_hbacsvcgroup',
            required=False,
            label=_(u'Member of HBAC service groups'),
        ),
    )


@register()
class hbacsvc_add(Method):
    __doc__ = _("Add a new HBAC service.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='service',
            label=_(u'Service name'),
            doc=_(u'HBAC service'),
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'HBAC service description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class hbacsvc_del(Method):
    __doc__ = _("Delete an existing HBAC service.")

    takes_args = (
        parameters.Str(
            'cn',
            multivalue=True,
            cli_name='service',
            label=_(u'Service name'),
            doc=_(u'HBAC service'),
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Flag(
            'continue',
            doc=_(u"Continuous mode: Don't stop on errors."),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Output(
            'result',
            dict,
            doc=_(u'List of deletions that failed'),
        ),
        output.ListOfPrimaryKeys(
            'value',
        ),
    )


@register()
class hbacsvc_find(Method):
    __doc__ = _("Search for HBAC services.")

    takes_args = (
        parameters.Str(
            'criteria',
            required=False,
            doc=_(u'A string searched in all relevant object attributes'),
        ),
    )
    takes_options = (
        parameters.Str(
            'cn',
            required=False,
            cli_name='service',
            label=_(u'Service name'),
            doc=_(u'HBAC service'),
            no_convert=True,
        ),
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'HBAC service description'),
        ),
        parameters.Int(
            'timelimit',
            required=False,
            label=_(u'Time Limit'),
            doc=_(u'Time limit of search in seconds (0 is unlimited)'),
        ),
        parameters.Int(
            'sizelimit',
            required=False,
            label=_(u'Size Limit'),
            doc=_(u'Maximum number of entries returned (0 is unlimited)'),
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'pkey_only',
            required=False,
            label=_(u'Primary key only'),
            doc=_(u'Results should contain primary key attribute only ("service")'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.ListOfEntries(
            'result',
        ),
        output.Output(
            'count',
            int,
            doc=_(u'Number of entries returned'),
        ),
        output.Output(
            'truncated',
            bool,
            doc=_(u'True if not all results were returned'),
        ),
    )


@register()
class hbacsvc_mod(Method):
    __doc__ = _("Modify an HBAC service.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='service',
            label=_(u'Service name'),
            doc=_(u'HBAC service'),
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Str(
            'description',
            required=False,
            cli_name='desc',
            label=_(u'Description'),
            doc=_(u'HBAC service description'),
        ),
        parameters.Str(
            'setattr',
            required=False,
            multivalue=True,
            doc=_(u'Set an attribute to a name/value pair. Format is attr=value.\nFor multi-valued attributes, the command replaces the values already present.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'addattr',
            required=False,
            multivalue=True,
            doc=_(u'Add an attribute/value pair. Format is attr=value. The attribute\nmust be part of the schema.'),
            exclude=('webui',),
        ),
        parameters.Str(
            'delattr',
            required=False,
            multivalue=True,
            doc=_(u'Delete an attribute/value pair. The option will be evaluated\nlast, after all sets and adds.'),
            exclude=('webui',),
        ),
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )


@register()
class hbacsvc_show(Method):
    __doc__ = _("Display information about an HBAC service.")

    takes_args = (
        parameters.Str(
            'cn',
            cli_name='service',
            label=_(u'Service name'),
            doc=_(u'HBAC service'),
            no_convert=True,
        ),
    )
    takes_options = (
        parameters.Flag(
            'rights',
            label=_(u'Rights'),
            doc=_(u'Display the access rights of this entry (requires --all). See ipa man page for details.'),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'all',
            doc=_(u'Retrieve and print all attributes from the server. Affects command output.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'raw',
            doc=_(u'Print entries as stored on the server. Only affects output format.'),
            exclude=('webui',),
            default=False,
            autofill=True,
        ),
        parameters.Flag(
            'no_members',
            doc=_(u'Suppress processing of membership attributes.'),
            exclude=('webui', 'cli'),
            default=False,
            autofill=True,
        ),
    )
    has_output = (
        output.Output(
            'summary',
            (unicode, type(None)),
            doc=_(u'User-friendly description of action performed'),
        ),
        output.Entry(
            'result',
        ),
        output.PrimaryKey(
            'value',
            doc=_(u"The primary_key value of the entry, e.g. 'jdoe' for a user"),
        ),
    )



#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#

"""
Module provides tests that uninstallation is successful.

It is important not to leave the remote system in an inconsistent
state. Every failed uninstall should successfully remove remaining
pieces if possible.
"""

from __future__ import absolute_import

import os

from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks
from ipaplatform.paths import paths
from ipaserver.install import dsinstance
from ipapython.ipaldap import realm_to_serverid


class TestUninstallBase(IntegrationTest):

    num_replicas = 1

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=False)

    def test_uninstall_client_invalid_hostname(self):

        # using replica as client just for convenience
        client = self.replicas[0]
        client_inv_hostname = '{}.nonexistent'.format(client.hostname)
        tasks.install_client(self.master, client,
                             extra_args=['--hostname', client_inv_hostname])

        client.run_command(['ipa-client-install', '--uninstall', '-U'])
        client_uninstall_log = client.get_file_contents(
            paths.IPACLIENT_UNINSTALL_LOG, encoding='utf-8'
        )
        assert "exception: ScriptError:" not in client_uninstall_log

        restore_state_path = os.path.join(paths.IPA_CLIENT_SYSRESTORE,
                                          'sysrestore.state')
        result = client.run_command(
            ['ls', restore_state_path], raiseonerr=False)
        assert 'No such file or directory' in result.stderr_text

    def test_install_uninstall_replica(self):
        # Test that the sequence install replica / uninstall replica
        # properly removes the line
        # Include /etc/httpd/conf.d/ipa-rewrite.conf
        # from ssl.conf on the replica
        tasks.install_replica(self.master, self.replicas[0],
                              extra_args=['--force-join'])
        tasks.uninstall_replica(self.master, self.replicas[0])
        errline = b'Include /etc/httpd/conf.d/ipa-rewrite.conf'
        ssl_conf = self.replicas[0].get_file_contents(paths.HTTPD_SSL_CONF)
        assert errline not in ssl_conf

    def test_failed_uninstall(self):
        self.master.run_command(['ipactl', 'stop'])

        serverid = realm_to_serverid(self.master.domain.realm)
        instance_name = ''.join([dsinstance.DS_INSTANCE_PREFIX, serverid])

        try:
            # Moving the DS instance out of the way will cause the
            # uninstaller to raise an exception and return with a
            # non-zero return code.
            self.master.run_command([
                '/bin/mv',
                '%s/%s' % (paths.ETC_DIRSRV, instance_name),
                '%s/%s.test' % (paths.ETC_DIRSRV, instance_name)
            ])

            cmd = self.master.run_command([
                'ipa-server-install',
                '--uninstall', '-U'],
                raiseonerr=False
            )
            assert cmd.returncode == 1
        finally:
            # Be paranoid. If something really went wrong then DS may
            # be marked as uninstalled so server cert will still be
            # tracked and the instances may remain. This can cause
            # subsequent installations to fail so be thorough.
            dashed_domain = self.master.domain.realm.replace(".", '-')
            dirsrv_service = "dirsrv@%s.service" % dashed_domain
            self.master.run_command(['systemctl', 'stop', dirsrv_service])

            # Moving it back should allow the uninstall to finish
            # successfully.
            self.master.run_command([
                '/bin/mv',
                '%s/%s.test' % (paths.ETC_DIRSRV, instance_name),
                '%s/%s' % (paths.ETC_DIRSRV, instance_name)
            ])

            cmd = self.master.run_command([
                'ipa-server-install',
                '--uninstall', '-U'],
                raiseonerr=False
            )
            assert cmd.returncode == 0

            self.master.run_command([
                paths.IPA_GETCERT,
                'stop-tracking',
                '-d', '%s/%s' % (paths.ETC_DIRSRV, instance_name),
                '-n', 'Server-Cert',
            ])

            self.master.run_command([
                paths.DSCTL, serverid, 'remove', '--do-it'
            ])
